#!/usr/bin/env python3

from sys import argv, stdin, version_info, version, executable
import json
import jedi
import inspect
import pydoc
import re
import keyword
import os

cr_version = int(os.getenv("CR_VERSION", "0"))

def is_identifier_char(s):
	if s == "": return False
	if hasattr(str, "isidentifier"): return ("_"+s).isidentifier()
	if s.isalnum(): return True
	s = s.replace("_", "")
	return s == "" or s.isalnum()

def standard_type(word):
	map = {
		"int": "int",
		"integer": "int",
		"str": "str",
		"string": "str",
		"dict": "dict",
		"dictionary": "dict",
		"number": "number",
		"bool": "bool",
		"boolean": "bool",
		"character": "str",
		"list": "list",
		"tuple": "tuple",
		"pair": "tuple",
		"set": "set",
		"true": "bool",
		"false": "bool",
		"-1": "int",
		"index": "int",
		"none": "None",
		"iterator": "iterator",
	}
	if word in map:
		return map[word]
	else:
		return None

def process_full_name(completion, item):
	full_name = ""
	name = completion.name
	try:
		full_name = completion.full_name
	except AttributeError:
		full_name = name
	if not full_name: full_name = name
	dot = full_name.rfind(".")
	if dot == -1:
		try:
			class_name = completion._name._obj.name
			if class_name:
				full_name = class_name + "." + full_name
		except:
			if full_name != name:
				dot = len(full_name)
				full_name = full_name + "." + name
	if dot != -1 and full_name[dot+1:] != name:
		t = standard_type(full_name[dot+1:])
		if t:
			item["type"] = t
		full_name = full_name[:dot+1] + name
		if item["kind"] == "class":
			item["kind"] = ""
	item["full_name"] = full_name

def insertion_for_params(params):
	if params == None: return "(<##>)"
	insertion = ""
	separator = False
	optional = 0
	for name in params:
		if len(name) == 0:
			continue
		name = name.replace(" = ", "=")
		search = name.find("=")
		ellipsis = False
		if optional or search != -1:
			if separator:
				insertion += " [, "
			else:
				insertion += "[ "
			optional += 1
		elif name == "...":
			ellipsis = True
			if separator:
				insertion += ", "
		elif separator:
			insertion += "#>, "
		if optional == 0 and separator and ellipsis == False:
			insertion += "<#"
		if separator == False:
			separator = True
		insertion += name
	if optional > 0:
		insertion += " "
	for x in range(0, optional):
		insertion += "]"
	if separator:
		insertion = "<#"+insertion+"#>"
	if insertion == "<#...#>":
		insertion = "<##>"
	insertion = "(" + insertion + ")"
	return insertion

def process_insertion(completion, item):
	if item["kind"] not in ["function", "method", "class"]:
		return
	params = None
	try:
		obj = pydoc.locate(item["full_name"], 0)
		if obj and callable(obj):
			signature = inspect.signature(obj)
			params = []
			for p in signature._parameters.values():
				params.append(str(p))
			if len(params) and params[0] == "self":
				if item["kind"] == "function" and item["full_name"].find(".") != -1:
					item["kind"] = "method"
				del params[0]
	except:
		params = None
	if params == None:
		try:
			params = [p.description for p in completion.params]
		except: pass
	item["insertion"] = insertion_for_params(params)

def docstring2html(docstring):
	while True:
		end = docstring.find("\n")
		if end == -1: break
		line = docstring[:end]
		if line.find(" -> ") != -1 or line.find(" --> ") != -1 or len(line) == 0 or line[-1] == ')':
			docstring = docstring[end+1:]
		else:
			break
	docstring = docstring.replace('<', '&lt;')
	docstring = docstring.replace('>', '&gt;')
	docstring = re.sub('``(.*?)(``|\'\')', '<code>\\1</code>', docstring)
	docstring = re.sub('`(.+?)(`|\')', '<code>\\1</code>', docstring)
	docstring = re.sub('\n:type\s+\w+\s*:\s*(.*?)(?=\n|$)', ' Type: <code>\\1</code>', docstring)
	docstring = re.sub('\n:param\s+(\w+)\s*:', '<br><b><code>\\1</code></b>:', docstring)
	docstring = re.sub(':rtype:', 'Return type: ', docstring)
	docstring = re.sub(':\w+:', '', docstring)
	docstring = re.sub(r'^\s*[=\-~*][=\-~*\s]*$', '<br>', docstring, flags=re.MULTILINE)
	docstring = docstring.replace("\n\n", "<br>")
	return docstring

cache = dict()
def cached_doc(name):
	if name in cache:
		return cache[name]
	doc = ""
	try:
		doc = pydoc.getdoc(pydoc.locate(name, 0)).strip()
	except: pass
	cache[name] = doc
	return doc

def process_description(completion, item):
	s = ""
	if item["kind"] != "module":
		try:
			object = pydoc.locate(item["full_name"], 0)
			if inspect.isroutine(object):
				s = pydoc.getdoc(object)
				if inspect.ismethoddescriptor(object):
					item["kind"] = "method"
				lines = s.splitlines()
				i = 0
				while i < len(lines):
					lines[i] = lines[i].strip()
					if len(lines[i]) and lines[i][0] == '|':
						del lines[i]
					else:
						i += 1
				s = "\n".join(lines)
		except:
			pass
	if not s:
		s = completion.docstring(raw=False)
	ignore_names = set()
	dot = item["full_name"].find(".")
	if dot != -1:
		class_name = item["full_name"][:dot]
		ignore_names.add(class_name)
	if "type" in item: ignore_names.add(item["type"])
	if item["string"] in ignore_names: ignore_names.remove(item["string"])
	for ignore_name in ignore_names:
		if not s: break
		ignore_doc = cached_doc(ignore_name)
		if ignore_doc:
			search = s.find(ignore_doc)
			if search != -1: s = s[:search]
	if item["full_name"] == "sys.version":
		if s: s += "\n\n"
		s += version + "\n\n" + executable
	item["plain_description"] = s
	s = docstring2html(s)
	if dot != -1:
		if cr_version > 50000:
			info = "<b><code>" + item["full_name"][:dot] + "</code></b>"
		else:
			info = "<code>" + item["full_name"][:dot] + "</code>"
		if s: s = info + " &ndash; " + s
	if s: item["description"] = s

def process_kind(completion, item):
	if item["kind"] == "class":
		try:
			obj = pydoc.locate(item["full_name"], 0)
			if issubclass(obj, BaseException):
				item["kind"] = "exception"
		except: pass
	elif item["kind"] == "statement":
		if item["string"].isupper():
			item["kind"] = "constant"
		else:
			try:
				obj = pydoc.locate(item["full_name"], 0)
				if inspect.ismethod(obj) or inspect.ismethoddescriptor(obj):
					item["kind"] = "method"
					return
				elif inspect.isfunction(obj) or callable(obj):
					item["kind"] = "function"
					return
			except: pass
			item["kind"] = "variable"

def guess_type(completion, item):
	if "plain_description" in item:
		types = set()
		description = item["plain_description"]
		for match in re.finditer(r'(?i)\b(return(s|ing)? )(a |an |the )?(\w+)', description):
			word = match.group(4).lower()
			t = standard_type(word)
			if t:
				types.add(t)
			else:
				types.clear()
				break
		if len(types) == 1:
			for x in types: return x
	return None

def process_type(completion, item):
	if (item["kind"] == "function" or item["kind"] == "method") and not item["string"].startswith("__"):
		map = {
			"builtins.len": "int",
			"builtins.abs": "number",
			"builtins.bin": "str",
			"builtins.callable": "bool",
			"builtins.chr": "str",
			"builtins.delattr": "None",
			"builtins.format": "str",
			"builtins.hasattr": "bool",
			"builtins.hash": "int",
			"builtins.hex": "str",
			"builtins.id": "int",
			"builtins.input": "str",
			"builtins.isinstance": "bool",
			"builtins.issubclass": "bool",
			"builtins.oct": "str",
			"builtins.ord": "int",
			"builtins.pow": "number",
			"builtins.print": "None",
			"builtins.repr": "str",
			"builtins.round": "number",
			"builtins.setattr": "None",
			"builtins.sorted": "list",
			"builtins.sum": "number",
			"builtins.ascii": "str",
			"builtins.dir": "list",
			"builtins.iter": "iterator",
			"builtins.vars": "dict",
			"list.append": "None",
			"float.as_integer_ratio": "tuple",
			"int.bit_length": "int",
			"str.capitalize": "str",
			"str.casefold": "str",
			"str.center": "str",
			"str.removeprefix": "str",
			"str.removesuffix": "str",
			"list.clear": "None",
			"int.conjugate": "int",
			"str.encode": "str",
			"int.as_integer_ratio": "tuple",
			"float.conjugate": "float",
			"str.expandtabs": "str",
			"list.extend": "None",
			"float.fromhex": "float",
			"dict.fromkeys": "dict",
			"float.hex": "str",
			"list.insert": "None",
			"str.join": "str",
			"str.ljust": "str",
			"str.lower": "str",
			"str.lstrip": "str",
			"str.maketrans": "dict",
			"str.partition": "tuple",
			"dict.popitem": "tuple",
			"list.remove": "None",
			"str.replace": "str",
			"list.reverse": "None",
			"str.rjust": "str",
			"str.rpartition": "tuple",
			"str.rstrip": "str",
			"str.strip": "str",
			"str.swapcase": "str",
			"str.title": "str",
			"int.to_bytes": "bytes",
			"str.translate": "str",
			"str.upper": "str",
			"str.zfill": "str",
			"list.copy": "list",
			"tuple.index": "int",
			"set.add": "None",
			"bool.as_integer_ratio": "tuple",
			"bool.bit_length": "str",
			"set.clear": "None",
			"bool.conjugate": "int",
			"set.copy": "set",
			"set.difference": "set",
			"set.difference_update": "None",
			"set.discard": "None",
			"set.intersection": "set",
			"set.intersection_update": "None",
			"set.issubset": "bool",
			"set.issuperset": "bool",
			"set.remove": "None",
			"set.symmetric_difference": "set",
			"set.symmetric_difference_update": "None",
			"bool.to_bytes": "bytes",
			"set.union": "set",
			"set.update": "None",
			"list.index": "int",
			"str.find": "int",
			"str.format": "str",
			"str.format_map": "str",
			"str.rfind": "int",
			"str.index": "int",
			"str.rindex": "int",
			"dict.copy": "dict",
		}
		if item["full_name"] in map:
			item["type"] = map[item["full_name"]]
		if not "type" in item:
			if "plain_description" in item:
				match = re.search(r'\) -?-> (\w+)( or None)?($|\.($|\s))', item["plain_description"])
				if match:
					item["type"] = match.group(1)
				else:
					t = guess_type(completion, item)
					if t: item["type"] = t

def get_position(source, line, column):
	source_lines = source.splitlines(True)
	if len(source_lines) == 0 or (len(source_lines[-1]) and source_lines[-1][-1] == "\n"):
		source_lines.append("")
	pos = 0
	valid = False
	for idx in range(line):
		if idx+1 == line:
			valid = len(source_lines[line-1].rstrip("\r\n")) >= column-1
			pos += column
			break
		if idx < len(source_lines):
			pos += len(source_lines[idx])
	if valid:
		return pos
	return -1

def process_completion(completion):
	name = completion.name
	item = {
		"string": name,
		"showImmediately": showImmediately,
		"docsUUID": docsUUID,
		"priority": 0 if name.startswith("__") else 1
	}
	kind = completion.type
	if isinstance(kind, str):
		item["kind"] = kind
	process_full_name(completion, item)
	process_kind(completion, item)
	process_insertion(completion, item)
	process_description(completion, item)
	process_type(completion, item)
	return item

filename = argv[1]
source = stdin.read()
if len(argv) > 2:
	line = int(argv[2])
	column = int(argv[3])-1
else:
	lines = source.splitlines()
	line = len(lines)
	column = len(lines[-1])
pos = get_position(source, line, column)
while pos > 0 and is_identifier_char(source[pos-1]):
	pos -= 1
	column -= 1
while pos < len(source) and is_identifier_char(source[pos]):
	source = source[:pos] + source[pos+1:]
script = jedi.Script(source, line, column, filename)
completions = script.completions()
if len(completions) == 0:
	# Jedi returns no completions if it doesn't know the object's type, which is unhelpful.
	# If completing after a dot, make Jedi return the set of completions for common builtin types.
	if pos != -1:
		while pos > 0 and source[pos-1].isspace(): pos -= 1
		if pos > 0 and source[pos-1] == '.':
			generic = "[list(),dict(),str(),int(),float(),tuple(),set()][_x_]."
			completing_chars = 0
			try:
				completing_chars = len(script._user_context.get_path_until_cursor())
			except:
				pass
			linestart = source.rfind("\n", 0, pos-completing_chars)+1
			match = re.match(r'\s*((\w+)\s+)$', source[linestart:pos-completing_chars])
			if match and keyword.iskeyword(match.group(2)):
				completing_chars += len(match.group(1))
			source = source[:pos-completing_chars] + generic + source[pos:]
			column += len(generic) - completing_chars
			script = jedi.Script(source, line, column, filename)
			completions = script.completions()
output = []
showImmediately = False
if len(argv) > 4:
	context = argv[4]
	if re.search(r'\.\s*$', context):
		if not re.search(r'\b\d\.\s*$', context):
			showImmediately = True
docsUUID = None
if version_info.major == 2:
	docsUUID = "64126A88-E29D-4B9D-AB53-13EF3E1D0388"
else:
	docsUUID = "FF667E76-BF73-4C18-AD70-EABAC0861D92"
for completion in completions:
	try:
		item = process_completion(completion)
		output.append(item)
		extra_completions = completion._same_name_completions
		if len(extra_completions):
			if item["kind"] in ["function", "method"] and item["full_name"].find(".") != -1 and not item["string"].startswith("__"):
				for extra_completion in extra_completions:
					if extra_completion in completions: continue
					try:
						extra_item = process_completion(extra_completion)
						if extra_item != item:
							output.append(extra_item)
					except:
						pass
	except:
		pass
print(json.dumps(output))
